/* --COPYRIGHT--,BSD
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

package com.demo.msp432blefirmwareupdater;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothManager;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanResult;
import android.bluetooth.le.ScanSettings;
import android.content.Context;
import android.content.Intent;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.support.v4.app.Fragment;
import android.support.v4.widget.SwipeRefreshLayout;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.Button;
import android.widget.ListView;
import android.widget.TextView;

import java.util.ArrayList;
import java.util.List;

public class BluetoothScanFragment extends Fragment {

    private static final String EXAMPLE_NAME = "exampleName";
    private static final String SCAN_FILTERS = "scanFilters";

    private String exampleName;

    private OnBluetoothScanFragmentListener mListener;

    private boolean connected;
    private BluetoothAdapter mBluetoothAdapter;
    private int REQUEST_ENABLE_BT = 1;
    private Handler mHandler;
    private Runnable mRunnable;
    private static final long SCAN_PERIOD = 10000;
    private BluetoothDevice btDevice;
    private BluetoothLeScanner mLEScanner;
    private ScanSettings settings;
    private List<ScanFilter> filters;
    private BluetoothGatt mGatt;
    private BluetoothDeviceListAdapter mBluetoothDeviceAdapter;
    private SwipeRefreshLayout mSwipeRefreshLayout;

    private ListView device_list;
    private Button connect_button;
    private Button disconnect_button;
    private TextView ble_scan_title;

    public BluetoothScanFragment() {
        // Required empty public constructor
    }

    public static BluetoothScanFragment newInstance(String example) {
        BluetoothScanFragment fragment = new BluetoothScanFragment();
        Bundle args = new Bundle();
        args.putString(EXAMPLE_NAME, example);
        fragment.setArguments(args);
        return fragment;
    }

    public static BluetoothScanFragment newInstance(ArrayList<ScanFilter> filters) {
        BluetoothScanFragment fragment = new BluetoothScanFragment();
        Bundle args = new Bundle();
        args.putSerializable(SCAN_FILTERS, filters);
        fragment.setArguments(args);
        return fragment;
    }

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setRetainInstance(true);

        if (savedInstanceState!= null) {

        }
        else {
            if (getArguments() != null) {
                exampleName = getArguments().getString(EXAMPLE_NAME);
                filters = (ArrayList<ScanFilter>) getArguments().getSerializable(SCAN_FILTERS);
            }

            connected = false;
        }
    }

    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
        // Inflate the layout for this fragment
        return inflater.inflate(R.layout.fragment_bluetooth_scan, container, false);
    }

    @Override
    public void onViewCreated(View view, Bundle savedInstanceState) {
        mHandler = new Handler();

        mSwipeRefreshLayout = (SwipeRefreshLayout)getActivity().findViewById(R.id.swiperefresh);

        device_list = (ListView) getActivity().findViewById(R.id.device_list);
        connect_button = (Button) getActivity().findViewById(R.id.connect_button);
        disconnect_button = (Button) getActivity().findViewById(R.id.disconnect_button);
        ble_scan_title = (TextView) getActivity().findViewById(R.id.ble_scan_title);

        mBluetoothDeviceAdapter = new BluetoothDeviceListAdapter(getActivity(), R.layout.ble_device_list_row, exampleName);
        device_list.setAdapter(mBluetoothDeviceAdapter);

        final BluetoothManager bluetoothManager =
                (BluetoothManager) getActivity().getSystemService(Context.BLUETOOTH_SERVICE);
        mBluetoothAdapter = bluetoothManager.getAdapter();

        mSwipeRefreshLayout.setOnRefreshListener(
                new SwipeRefreshLayout.OnRefreshListener() {
                    @Override
                    public void onRefresh() {
                        if (connected) {
                            mSwipeRefreshLayout.setRefreshing(false);
                            mSwipeRefreshLayout.destroyDrawingCache();
                            mSwipeRefreshLayout.clearAnimation();
                        }
                        else
                            startScan();
                    }
                }
        );

        if (connected) {
            connect_button.setVisibility(View.GONE);
            disconnect_button.setVisibility(View.VISIBLE);
        }
        else {
            connect_button.setVisibility(View.VISIBLE);
            disconnect_button.setVisibility(View.GONE);
        }

        connect_button.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                if (mListener != null && mBluetoothDeviceAdapter.getSelectedDevice() != null)
                    mListener.connectToBluetoothDevice(mBluetoothDeviceAdapter.getSelectedDevice());
            }
        });

        disconnect_button.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                if (mListener != null)
                    mListener.connectToBluetoothDevice(null);
            }
        });


        device_list.setOnItemClickListener(new AdapterView.OnItemClickListener() {
            @Override
            public void onItemClick(AdapterView<?> adapterView, View view, int position, long l) {
                getActivity().runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        mHandler.removeCallbacks(mRunnable);
                        mSwipeRefreshLayout.setRefreshing(false);
                        mSwipeRefreshLayout.destroyDrawingCache();
                        mSwipeRefreshLayout.clearAnimation();
                        scanLeDevice(false);
                    }
                });
                btDevice = ((BluetoothDevice) (adapterView.getItemAtPosition(position)));

                mBluetoothDeviceAdapter.setSelectedDevice(btDevice);
                mBluetoothDeviceAdapter.notifyDataSetChanged();

                String clicked = btDevice.getName();
                if (clicked == null)
                    clicked = "Unknown Device";
                Log.d("Selected ", clicked);
            }
        });
    }

    @Override
    public void onResume() {
        super.onResume();
        mHandler.removeCallbacks(mRunnable);
        mSwipeRefreshLayout.setRefreshing(false);
        if (mLEScanner != null && mBluetoothAdapter.isEnabled())
            mLEScanner.stopScan(mScanCallback);
        Log.d("BluetoothScanFragment", "onResume()");

        if (!connected)
            startScan();
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
        Log.d("BluetoothScanFragment", "onDestroy()");

        if (mGatt != null) {
            mGatt.close();
            mGatt = null;
        }
    }

    @Override
    public void onAttach(Context context) {
        super.onAttach(context);
        if (context instanceof OnBluetoothScanFragmentListener) {
            mListener = (OnBluetoothScanFragmentListener) context;
        } else {
            throw new RuntimeException(context.toString()
                    + " must implement OnBluetoothScanFragmentListener");
        }
    }

    @Override
    public void onDetach() {
        super.onDetach();
        mListener = null;
    }

    private void startScan() {
        mSwipeRefreshLayout.setRefreshing(true);
        if (mBluetoothAdapter == null || !mBluetoothAdapter.isEnabled()) {
            Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT);
        } else {
            if (Build.VERSION.SDK_INT >= 21) {
                mLEScanner = mBluetoothAdapter.getBluetoothLeScanner();
                settings = new ScanSettings.Builder()
                        .setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY)
                        .build();
                filters = new ArrayList<ScanFilter>();
            }

            mBluetoothDeviceAdapter = new BluetoothDeviceListAdapter(getActivity(), R.layout.ble_device_list_row, exampleName);
            device_list.setAdapter(mBluetoothDeviceAdapter);

            scanLeDevice(true);
        }
    }

    private void scanLeDevice(final boolean enable) {
        if (enable) {
            mRunnable = new Runnable() {
                @Override
                public void run() {
                    if (mLEScanner != null && mBluetoothAdapter.isEnabled())
                        mLEScanner.stopScan(mScanCallback);
                    mSwipeRefreshLayout.setRefreshing(false);
                }
            };
            mHandler.postDelayed(mRunnable, SCAN_PERIOD);
            mLEScanner.startScan(filters, settings, mScanCallback);
        } else {
            mSwipeRefreshLayout.setRefreshing(false);
            mLEScanner.stopScan(mScanCallback);
        }
    }

    private ScanCallback mScanCallback = new ScanCallback() {
        @Override
        public void onScanResult(int callbackType, ScanResult result) {
            final BluetoothDevice btDevice = result.getDevice();
            final int rssi = result.getRssi();

            mBluetoothDeviceAdapter.addDevice(btDevice, rssi);
            mBluetoothDeviceAdapter.notifyDataSetChanged();
        }

        @Override
        public void onBatchScanResults(List<ScanResult> results) {
            for (ScanResult sr : results) {
            }
        }

        @Override
        public void onScanFailed(int errorCode) {
            Log.e("Scan Failed", "Error Code: " + errorCode);
        }
    };

    public void deviceConnected(final BluetoothDevice device) {
        getActivity().runOnUiThread(new Runnable() {
            @Override
            public void run() {
                connect_button.setVisibility(View.GONE);
                disconnect_button.setVisibility(View.VISIBLE);
                connected = true;
                ble_scan_title.setText("Connected: " + device.getName());
            }
        });
    }

    public void deviceDisconnected() {
        getActivity().runOnUiThread(new Runnable() {
            @Override
            public void run() {
                connect_button.setVisibility(View.VISIBLE);
                disconnect_button.setVisibility(View.GONE);
                connected = false;
                ble_scan_title.setText("Select Bluetooth Device");
            }
        });
    }

    public interface OnBluetoothScanFragmentListener {
        void connectToBluetoothDevice(BluetoothDevice btdevice);
    }
}
