/* --COPYRIGHT--,BSD
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

package com.demo.msp432blefirmwareupdater;

import android.Manifest;
import android.bluetooth.BluetoothDevice;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.net.Uri;
import android.support.v4.app.ActivityCompat;
import android.support.v4.app.FragmentManager;
import android.support.v4.app.FragmentTransaction;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.widget.Button;

import com.demo.msp432blefirmwareupdater.gatt.GattManager;

public class MainActivity extends AppCompatActivity
        implements BluetoothScanFragment.OnBluetoothScanFragmentListener, OADDownloadFragment.OnFragmentInteractionListener, OADFragment.OnFragmentInteractionListener {

    private static final String BT_SCAN_TAG = "BT_SCAN_TAG";
    private static final String OAD_TAG = "OAD_TAG";
    private static final String OAD_DOWNLOAD_TAG = "OAD_DOWNLOAD_TAG";

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        int permissionCheck = ActivityCompat.checkSelfPermission(this,
                Manifest.permission.ACCESS_COARSE_LOCATION);
        if (permissionCheck == PackageManager.PERMISSION_DENIED) {
            int MY_PERMISSIONS_REQUEST_ACCESS_COARSE_LOCATION = 1;
            ActivityCompat.requestPermissions(this,
                    new String[]{Manifest.permission.ACCESS_COARSE_LOCATION},
                    MY_PERMISSIONS_REQUEST_ACCESS_COARSE_LOCATION);
        }

        // Setup Privacy Policy button
        Button privacy_policy = (Button) findViewById(R.id.privacy_policy);
        privacy_policy.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                String url = "http://www.ti.com/corp/docs/legal/privacy.shtml";

                Intent i = new Intent(Intent.ACTION_VIEW);
                i.setData(Uri.parse(url));
                startActivity(i);
            }
        });

        // Setup Terms of Use button
        Button terms_of_use = (Button) findViewById(R.id.terms_of_use);
        terms_of_use.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                String url = "http://www.ti.com/corp/docs/legal/termsofuse.shtml?DCMP=MISCELLANEOUS+Other&HQS=Other+OT+termsofuse";

                Intent i = new Intent(Intent.ACTION_VIEW);
                i.setData(Uri.parse(url));
                startActivity(i);
            }
        });

        Button program_button = findViewById(R.id.program_button);
        program_button.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                OADFragment oadFragment = (OADFragment) getSupportFragmentManager().findFragmentByTag(OAD_TAG);
                if (oadFragment != null)
                    oadFragment.beginFirmwareUpdate();
            }
        });

        if (savedInstanceState == null) {
            // Create a new Fragment to be placed in the activity layout
            OADFragment oadFragment = OADFragment.newInstance(null);
            BluetoothScanFragment bluetoothScanFragment = BluetoothScanFragment.newInstance("MSP432 OAD");

            FragmentManager frag_manager = getSupportFragmentManager();
            FragmentTransaction frag_transaction = frag_manager.beginTransaction();

            frag_transaction.add(R.id.fragment_container_scan, bluetoothScanFragment, BT_SCAN_TAG);
            frag_transaction.add(R.id.fragment_container_oad, oadFragment, OAD_TAG);

            frag_transaction.commit();

        }
        else {
            OADDownloadFragment oadDownloadFragment = (OADDownloadFragment) getSupportFragmentManager().findFragmentByTag(OAD_DOWNLOAD_TAG);
            if (oadDownloadFragment != null) {
                findViewById(R.id.ble_firmware_updater).setVisibility(View.GONE);
            }
        }
    }

    public void connectToBluetoothDevice(BluetoothDevice btdevice) {
        if (btdevice != null)
            Log.d("OADActivity", "Device selected: " + btdevice.getName());
        if (findViewById(R.id.fragment_container_scan) != null) {
            OADFragment oadFragment = (OADFragment) getSupportFragmentManager().findFragmentByTag(OAD_TAG);
            if (oadFragment != null)
                oadFragment.connectToDevice(btdevice);
        }
    }

    @Override
    public void BeginOADDownload(BluetoothDevice btdevice, GattManager gattManager, Uri selectedFile, String selectedFileType, byte firmwareTarget, String firmwareVersion, String firmwareAddress) {
        Log.d("OADActivity", "Begin OAD Download Fragment");
        if (findViewById(R.id.fragment_main) != null) {

            findViewById(R.id.ble_firmware_updater).setVisibility(View.GONE);

            // Create a new Fragment to be placed in the activity layout
            OADDownloadFragment oadDownloadFragment = OADDownloadFragment.newInstance(btdevice, selectedFile, selectedFileType, firmwareTarget, firmwareVersion, firmwareAddress);
            oadDownloadFragment.setGattManager(gattManager);

            FragmentTransaction transaction = getSupportFragmentManager().beginTransaction();

            transaction.replace(R.id.fragment_main, oadDownloadFragment, OAD_DOWNLOAD_TAG);
            transaction.detach(getSupportFragmentManager().findFragmentByTag(OAD_TAG));
            transaction.detach(getSupportFragmentManager().findFragmentByTag(BT_SCAN_TAG));
            transaction.addToBackStack(null);

            transaction.commit();
        }
    }

    @Override
    public void OADDownloadCompleted(boolean completed) {
        getSupportFragmentManager().popBackStack();
        findViewById(R.id.ble_firmware_updater).setVisibility(View.VISIBLE);
    }

    @Override
    public void onDeviceConnected(BluetoothDevice device) {
        BluetoothScanFragment bluetoothScanFragment = (BluetoothScanFragment) getSupportFragmentManager().findFragmentByTag(BT_SCAN_TAG);
        if (bluetoothScanFragment != null)
            bluetoothScanFragment.deviceConnected(device);
    }

    @Override
    public void onDeviceDisconnected() {
        BluetoothScanFragment bluetoothScanFragment = (BluetoothScanFragment) getSupportFragmentManager().findFragmentByTag(BT_SCAN_TAG);
        if (bluetoothScanFragment != null)
            bluetoothScanFragment.deviceDisconnected();

        findViewById(R.id.ble_firmware_updater).setVisibility(View.VISIBLE);
    }
}
