/* --COPYRIGHT--,BSD
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

package com.demo.msp432blefirmwareupdater;

import android.app.Activity;
import android.app.ProgressDialog;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothGattService;
import android.bluetooth.BluetoothProfile;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.provider.OpenableColumns;
import android.support.v4.app.Fragment;
import android.support.v7.app.AlertDialog;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Switch;
import android.widget.TextView;

import com.demo.msp432blefirmwareupdater.gatt.CharacteristicChangeListener;
import com.demo.msp432blefirmwareupdater.gatt.GattManager;
import com.demo.msp432blefirmwareupdater.gatt.GattOperationBundle;
import com.demo.msp432blefirmwareupdater.gatt.operations.GattCharacteristicReadOperation;
import com.demo.msp432blefirmwareupdater.gatt.operations.GattCharacteristicWriteOperation;
import com.demo.msp432blefirmwareupdater.gatt.operations.GattDescriptorReadOperation;
import com.demo.msp432blefirmwareupdater.gatt.operations.GattOperation;
import com.demo.msp432blefirmwareupdater.gatt.operations.GattSetNotificationOperation;

import java.util.List;
import java.util.UUID;

import static android.bluetooth.BluetoothGatt.GATT_SUCCESS;

public class OADFragment extends Fragment {

    private final String OAD_SERVICE_UUID = "f000ffc0-0451-4000-b000-000000000000";
    private final String OAD_IMG_IDENTIFY_CHAR_UUID = "f000ffc1-0451-4000-b000-000000000000";
    private final String OAD_IMG_BLOCK_CHAR_UUID = "f000ffc2-0451-4000-b000-000000000000";
    private final String OAD_EXT_CTL_CHAR_UUID = "f000ffc5-0451-4000-b000-000000000000";

    private static final String BT_DEVICE = "bt_device";

    private BluetoothDevice mBtdevice;
    private GattManager mGattManager;

    private TextView selected_firmware_textview;
    private Button firmware_select_button;
    private EditText fw_version_edittext;
    private EditText fw_address_edittext;
    private Button begin_update_button;
    private Switch sap_snp_switch;

    private static final int READ_REQUEST_CODE = 42;

    private final byte FIRMWARE_CC2650SNP = 0x03;
    private final byte FIRMWARE_MSP432SAP = 0x01;

    private Uri selectedFile;
    private String selectedFileExtension;

    private ProgressDialog mDialog;
    private Handler mHandler;
    private Runnable mRunnable;
    private boolean hasService;

    OnFragmentInteractionListener mListener;

    public OADFragment() {
        // Required empty public constructor
    }

    public static OADFragment newInstance(BluetoothDevice btdevice) {
        OADFragment fragment = new OADFragment();
        Bundle args = new Bundle();
        args.putParcelable(BT_DEVICE, btdevice);
        fragment.setArguments(args);
        return fragment;
    }

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setRetainInstance(true);


        if (savedInstanceState!= null) {

        }
        else {
            if (getArguments() != null) {
                mBtdevice = getArguments().getParcelable(BT_DEVICE);
            }
        }
    }

    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
        // Inflate the layout for this fragment
        return inflater.inflate(R.layout.fragment_oad, container, false);
    }

    @Override
    public void onViewCreated(View view, Bundle savedInstanceState) {
        sap_snp_switch= (Switch) getActivity().findViewById(R.id.sap_snp_switch);
        selected_firmware_textview = (TextView) getActivity().findViewById(R.id.selected_firmware_textview);
        firmware_select_button = (Button) getActivity().findViewById(R.id.firmware_select_button);

        firmware_select_button.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                Log.d("OADFragment", "Firmware select button clicked");
                performFileSearch();
            }
        });

        fw_version_edittext = (EditText) getActivity().findViewById(R.id.fw_version_edittextt);
        fw_address_edittext = (EditText) getActivity().findViewById(R.id.fw_address_edittextt);
    }

    @Override
    public void onResume() {
        super.onResume();
//        ((AppCompatActivity) getActivity()).getSupportActionBar().setTitle("MSP432 OAD");

        if (selectedFile != null) {
            updateSelectedFirmware(selectedFile);
        }
    }

    @Override
    public void onDestroy() {
        Log.d("OADFragment", "OnDestroy");
        super.onDestroy();
        if (mGattManager!= null) {
            mGattManager.close(mBtdevice);
            mGattManager = null;
        }
    }

    @Override
    public void onAttach(Context context) {
        super.onAttach(context);
        if (context instanceof OnFragmentInteractionListener) {
            mListener = (OnFragmentInteractionListener) context;
        } else {
            throw new RuntimeException(context.toString()
                    + " must implement OnFragmentInteractionListener");
        }
    }

    @Override
    public void onDetach() {
        super.onDetach();
        mListener = null;
    }

    public interface OnFragmentInteractionListener {
        void BeginOADDownload(BluetoothDevice btdevice, GattManager gattManager, Uri selectedFile, String selectedFileType, byte firmwareTarget, String firmwareVersion, String firmwareAddress);
        void onDeviceConnected(BluetoothDevice device);
        void onDeviceDisconnected();
    }

    public void connectToDevice(final BluetoothDevice device) {
        hasService = false;

        if (device != null) {
            final BluetoothGattCallback gattCallback = new BluetoothGattCallback() {
                @Override
                public void onConnectionStateChange(final BluetoothGatt gatt, int status, int newState) {
                    super.onConnectionStateChange(gatt, status, newState);

                    if (status == 133) {
                        Log.e("GattManager", "Got the status 133 bug, closing gatt");
                        gatt.close();
                        mGattManager.getGatts().remove(mBtdevice.getAddress());
                        if (mListener != null)
                            mListener.onDeviceDisconnected();
                        return;
                    }

                    if (newState == BluetoothProfile.STATE_CONNECTED) {
                        Log.i("GattManager", "Gatt connected to device " + mBtdevice.getAddress());
                        getActivity().runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                final Handler handler = new Handler();
                                handler.postDelayed(new Runnable() {
                                    @Override
                                    public void run() {
                                        // Do something after 1000ms
                                        gatt.discoverServices();
                                    }
                                }, 1000);
                            }
                        });
                    } else if (newState == BluetoothProfile.STATE_DISCONNECTED) {
                        Log.i("GattManager", "Disconnected from gatt server " + mBtdevice.getAddress() + ", newState: " + newState);
                        mGattManager.getGatts().remove(mBtdevice.getAddress());
                        mGattManager.setCurrentOperation(null);
                        gatt.close();
                        mGattManager.drive();
                        mBtdevice = null;
                        mGattManager = null;

                        getActivity().runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                mDialog.dismiss();
                                mHandler.removeCallbacks(mRunnable);
                                getActivity().runOnUiThread(new Runnable() {
                                    @Override
                                    public void run() {
                                        returnToDeviceSelection("Device disconnected.");
                                    }
                                });
                            }
                        });

                        if (mListener != null)
                            mListener.onDeviceDisconnected();
                    }
                }

                @Override
                public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                    super.onDescriptorRead(gatt, descriptor, status);
                    ((GattDescriptorReadOperation) mGattManager.getCurrentOperation()).onRead(descriptor);
                    mGattManager.setCurrentOperation(null);
                    mGattManager.drive();
                }

                @Override
                public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                    super.onDescriptorWrite(gatt, descriptor, status);
                    mGattManager.setCurrentOperation(null);
                    mGattManager.drive();
                }

                @Override
                public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                    super.onCharacteristicRead(gatt, characteristic, status);
                    ((GattCharacteristicReadOperation) mGattManager.getCurrentOperation()).onRead(characteristic);
                    mGattManager.setCurrentOperation(null);
                    mGattManager.drive();
                }

                @Override
                public void onServicesDiscovered(final BluetoothGatt gatt, int status) {
                    super.onServicesDiscovered(gatt, status);
                    Log.d("GattManager", "services discovered, status: " + status);

                    getActivity().runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            mDialog.dismiss();
                            mHandler.removeCallbacks(mRunnable);
                        }
                    });

                    List<BluetoothGattService> services = gatt.getServices();

                    for (BluetoothGattService s : services) {
                        if (s.getUuid().toString().equals(OAD_SERVICE_UUID)) {
                            hasService = true;
                        }
                    }
                    if (!hasService) {
                        getActivity().runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                gatt.close();
                                returnToDeviceSelection("Selected device does not contain the expected BLE services/characteristics. Make sure correct firmware is programmed and the appropriate device is being selected.");
                            }
                        });
                    }
                    else {
                        mGattManager.getGatts().put(mBtdevice.getAddress(), gatt);
                        gatt.requestMtu(512);

                        if (mListener != null)
                            mListener.onDeviceConnected(mBtdevice);

                    }
                }

                @Override
                public void onMtuChanged(BluetoothGatt gatt, int mtu, int status) {
                    if (status == GATT_SUCCESS)
                        Log.d("GattManager", "MTU successfully changed to: " + mtu);

                    GattOperationBundle bundle = new GattOperationBundle();

                    bundle.addOperation(new GattSetNotificationOperation(
                            mBtdevice,
                            UUID.fromString(OAD_SERVICE_UUID),
                            UUID.fromString(OAD_IMG_IDENTIFY_CHAR_UUID),
                            UUID.fromString("00002902-0000-1000-8000-00805f9b34fb"),
                            true
                    ));

                    bundle.addOperation(new GattSetNotificationOperation(
                            mBtdevice,
                            UUID.fromString(OAD_SERVICE_UUID),
                            UUID.fromString(OAD_IMG_BLOCK_CHAR_UUID),
                            UUID.fromString("00002902-0000-1000-8000-00805f9b34fb"),
                            true
                    ));

                    bundle.addOperation(new GattSetNotificationOperation(
                            mBtdevice,
                            UUID.fromString(OAD_SERVICE_UUID),
                            UUID.fromString(OAD_EXT_CTL_CHAR_UUID),
                            UUID.fromString("00002902-0000-1000-8000-00805f9b34fb"),
                            true
                    ));

                    mGattManager.queue(bundle);
                }

                @Override
                public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                    super.onCharacteristicWrite(gatt, characteristic, status);
                    Log.d("GattManager", "Characteristic " + characteristic.getUuid() + "written to on device " + mBtdevice.getAddress());
                    if (mGattManager.getCurrentOperation() != null && mGattManager.getCurrentOperation().type() == GattOperation.OperationType.OPERATION_CHAR_WRITE)
                        ((GattCharacteristicWriteOperation) mGattManager.getCurrentOperation()).onWrite(characteristic, status);
                    mGattManager.setCurrentOperation(null);
                    mGattManager.drive();
                }

                @Override
                public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
                    super.onCharacteristicChanged(gatt, characteristic);
                    Log.d("GattManager", "Characteristic " + characteristic.getUuid() + "was changed, device: " + device.getAddress());
                    if (mGattManager.getCharacteristicChangeListeners().containsKey(characteristic.getUuid())) {
                        for (CharacteristicChangeListener listener : mGattManager.getCharacteristicChangeListeners().get(characteristic.getUuid())) {
                            listener.onCharacteristicChanged(mBtdevice.getAddress(), characteristic);
                        }
                    }
                }
            };

            mBtdevice = device;
            mGattManager = new GattManager(getActivity());
            mGattManager.connectToDevice(mBtdevice, gattCallback);

            mDialog = ProgressDialog.show(getActivity(), "", "Connecting...", true);
            mHandler = new android.os.Handler();
            mRunnable = new Runnable() {
                public void run() {
                    Log.i("tag", "10000 milliseconds BLE Connection timeout");
                    if (!hasService) {
                        getActivity().runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                mDialog.dismiss();
                                mHandler.removeCallbacks(mRunnable);
                                returnToDeviceSelection("BLE Connection Timed out");
                            }
                        });
                    }
                }
            };
            mHandler.postDelayed(mRunnable, 10000);

        }
        else {
            if (mGattManager!= null) {
                mGattManager.close(mBtdevice);
//                mGattManager = null;
            }
        }
    }

    public void beginFirmwareUpdate() {
        if (mBtdevice != null && mGattManager != null && mListener != null && selectedFile != null) {
            if (selectedFileExtension.equals("txt") || selectedFileExtension.equals("hex")) {
                byte firmwareTarget;
                if (sap_snp_switch.isChecked())
                    firmwareTarget = FIRMWARE_CC2650SNP;
                else
                    firmwareTarget = FIRMWARE_MSP432SAP;

                String fw_version = fw_version_edittext.getText().toString();
                String fw_address = fw_address_edittext.getText().toString();
                // Start OAD Download Fragment
                mListener.BeginOADDownload(mBtdevice, mGattManager, selectedFile, selectedFileExtension, firmwareTarget, fw_version, fw_address);
            }
        }
    }

    public void performFileSearch() {

        // ACTION_OPEN_DOCUMENT is the intent to choose a file via the system's file
        // browser.
        Intent intent = new Intent(Intent.ACTION_GET_CONTENT);

        // Filter to only show results that can be "opened", such as a
        // file (as opposed to a list of contacts or timezones)
        intent.addCategory(Intent.CATEGORY_OPENABLE);

        // Filter to show only images, using the image MIME data type.
        // If one wanted to search for ogg vorbis files, the type would be "audio/ogg".
        // To search for all documents available via installed storage providers,
        // it would be "*/*".
        intent.setType("*/*");

        String[] mimeTypes = {"text/plain", "application/octet-stream"};
        intent .putExtra(Intent.EXTRA_MIME_TYPES, mimeTypes);

        startActivityForResult(intent, READ_REQUEST_CODE);
    }

    @Override
    public void onActivityResult(int requestCode, int resultCode,
                                 Intent resultData) {

        // The ACTION_OPEN_DOCUMENT intent was sent with the request code
        // READ_REQUEST_CODE. If the request code seen here doesn't match, it's the
        // response to some other intent, and the code below shouldn't run at all.

        if (requestCode == READ_REQUEST_CODE && resultCode == Activity.RESULT_OK) {
            // The document selected by the user won't be returned in the intent.
            // Instead, a URI to that document will be contained in the return intent
            // provided to this method as a parameter.
            // Pull that URI using resultData.getData().
            Uri uri = null;
            if (resultData != null) {
                uri = resultData.getData();
                Log.i("OADFragment", "Uri: " + uri.toString());
                String extension = updateSelectedFirmware(uri);
                if (extension.equals("txt") || extension.equals("hex")) {
                    selectedFile = uri;
                    selectedFileExtension = extension;
                }
                else {
                    Log.e("OADFragment", "Incorrect file type selected. Must be .txt or .hex");
                }
            }
        }
    }

    public String updateSelectedFirmware(Uri uri) {
        // The query, since it only applies to a single document, will only return
        // one row. There's no need to filter, sort, or select fields, since we want
        // all fields for one document.
        Cursor cursor = getActivity().getContentResolver()
                .query(uri, null, null, null, null, null);

        try {
            // moveToFirst() returns false if the cursor has 0 rows.  Very handy for
            // "if there's anything to look at, look at it" conditionals.
            if (cursor != null && cursor.moveToFirst()) {

                // Note it's called "Display Name".  This is
                // provider-specific, and might not necessarily be the file name.
                String displayName = cursor.getString(
                        cursor.getColumnIndex(OpenableColumns.DISPLAY_NAME));
                Log.i("OADFragment", "Display Name: " + displayName);

                int sizeIndex = cursor.getColumnIndex(OpenableColumns.SIZE);
                // If the size is unknown, the value stored is null.  But since an
                // int can't be null in Java, the behavior is implementation-specific,
                // which is just a fancy term for "unpredictable".  So as
                // a rule, check if it's null before assigning to an int.  This will
                // happen often:  The storage API allows for remote files, whose
                // size might not be locally known.
                String size = null;
                if (!cursor.isNull(sizeIndex)) {
                    // Technically the column stores an int, but cursor.getString()
                    // will do the conversion automatically.
                    size = cursor.getString(sizeIndex);
                } else {
                    size = "Unknown";
                }
                Log.i("OADFragment", "Size: " + size);

                String extension = "";

                if (displayName.lastIndexOf(".") >= 0)
                    extension = displayName.substring(displayName.lastIndexOf(".")+1).toLowerCase();

                Log.i("OADFragment", "File Extension: " + extension);

                if (extension.equals("txt") || extension.equals("hex"))
                    selected_firmware_textview.setText(displayName + " - " + size + " bytes");

                return extension;
            }
            return "";
        } finally {
            cursor.close();
        }
    }

    private void returnToDeviceSelection(String msg) {
        AlertDialog.Builder builder = new AlertDialog.Builder(getActivity());
        builder.setMessage(msg)
                .setCancelable(false)
                .setPositiveButton("OK", new DialogInterface.OnClickListener() {
                    public void onClick(DialogInterface dialog, int id) {
                        getActivity().getSupportFragmentManager().popBackStack();
                    }
                });
        AlertDialog alert = builder.create();
        alert.show();
    }

}
